--[[
Make players into monsters.
]]

-- TODO: aim down shallow cliffs
-- TODO: trigger item / monster activations


-- preamble
Triggers = {}  -- eliminate "compatibility" triggers


-- constants
berserk_threshold = 150/2  -- health threshold at which players berserk
cooldown_delta = 60  -- time to wait after firing
firing_delta = 30  -- time for which firing is allowed
travel_speed = 0.03  -- WU/tick walking travel speed
berserk_travel_speed = 0.07  -- WU/tick running travel speed
player_height = 1.0  -- amount in WU of required headclearance
player_thiccness = 0.26  -- amount in WU required gutclearance
player_step = 0.25  -- amount in WU that the player can step up
pitch_step = 1.0  -- pitch adjustment rate in degrees per tick
yaw_step = 3.0  -- yaw adjustment rate in degrees per tick


-- globals
next_lines = {}  -- polygon index -> polygon index -> nil | true | line

--[[
extra fields on player objects:
  _target: pointer to the object the player is currently tracking, if any
  _cooldown_timer: 0 = no timer; pos = time 'til which the player can fire; neg = -(time at which the player has recharged)
  _track_items: if true, track items
  _track_monsters: if true, track monsters
  _autofire: if true, aim and fire automatically
  _old_position: optional {x, y, z, polygon}, stashed player position from previous tick
]]


--[[
Precalculates the `next_lines` table, used for monster pathfinding.

Based on Wikipedia's article on Floyd-Warshall.
]]
function precalculate_paths()
    local distances = {}  -- polygon -> polygon -> int distance
    local i, j, k, line
    local polygon_count, line_count = #Polygons, #Lines

    next_lines = {}  -- polygon -> polygon -> line or nil
    distances[polygon_count] = false  -- initialize outer array size
    next_lines[polygon_count] = false  -- initialize outer array size

    for i = 0, polygon_count, 1 do
        distances[i] = {}
        distances[i][polygon_count] = false  -- initialize inner array size
        for j = 0, polygon_count, 1 do
            distances[i][j] = 2 + polygon_count  --  a/k/a infinity
        end
        next_lines[i] = {}
        next_lines[i][polygon_count] = false  -- initialize inner array size
        next_lines[i][i] = true
        distances[i][i] = 0
    end

    for line in Lines() do
        local cw_index = line.cw_polygon and line.cw_polygon.index
        local ccw_index = line.ccw_polygon and line.ccw_polygon.index
        if passablep(line, "cw_to_ccw") then
            next_lines[cw_index][ccw_index] = line
            distances[cw_index][ccw_index] = 1
        end
        if passablep(line, "ccw_to_cw") then
            next_lines[ccw_index][cw_index] = line
            distances[ccw_index][cw_index] = 1
        end
    end

    for k = 0, polygon_count, 1 do
        for i = 0, polygon_count, 1 do
            local distances_ik = distances[i][k]
            local next_lines_ik = next_lines[i][k]
            for j = 0, polygon_count, 1 do
                local distances_kj = distances[k][j]
                if distances[i][j] > distances_ik + distances_kj then
                    distances[i][j] = distances_ik + distances_kj
                    next_lines[i][j] = next_lines_ik
                end
            end
        end
    end
end


--[[
Test the passability of a line in one direction.
]]
-- TODO: handle passable sides
-- TODO: handle line width
function passablep(line, direction)
    local source_polygon, target_polygon
    if direction == "cw_to_ccw" then
        source_polygon = line.cw_polygon
        target_polygon = line.ccw_polygon
    else
        source_polygon = line.ccw_polygon
        target_polygon = line.cw_polygon
    end

    if not source_polygon or not target_polygon then
        return nil
    end

    -- check that the window can be wide enough
    local source_ceiling, target_ceiling, source_floor, target_floor
    source_ceiling = (
        (source_polygon.platform and source_polygon.platform.maximum_ceiling_height)
        or source_polygon.ceiling.height
    )
    target_ceiling = (
        (target_polygon.platform and target_polygon.platform.maximum_ceiling_height)
        or target_polygon.ceiling.height
    )
    source_floor = (
        (source_polygon.platform and source_polygon.platform.minimum_floor_height)
        or source_polygon.floor.height
    )
    target_floor = (
        (target_polygon.platform and target_polygon.platform.minimum_floor_height)
        or target_polygon.floor.height
    )
    if (math.min(source_ceiling, target_ceiling) - math.max(source_floor, target_floor) <
            player_height) then
        return false
    end

    -- check that the player can climb from one floor to the other
    local source_floor = (
        (source_polygon.platform and source_polygon.platform.maximum_floor_height)
        or source_polygon.floor.height
    )
    local target_floor = (
        (target_polygon.platform and target_polygon.platform.minimum_floor_height)
        or target_polygon.floor.height
    )
    if (source_polygon.media  -- handle swimming
            and source_polygon.media.type ~= "lava"
            and source_polygon.media.type ~= "goo") then
        source_floor = math.max(
            source_floor,
            source_polygon.media.high - player_step
        )
    end

    return (source_floor + player_step >= target_floor)
end


--[[
Predicate for checking whether two players are friends.
]]
function friendp(victim, aggressor)  -- Player, Monster --> bool
    if is_player(aggressor) then
        return ((victim.monster == aggressor) or
            (victim.life > berserk_threshold and victim.team == aggressor.team))
    elseif is_monster(aggressor) then
        return (victim.life > berserk_threshold and
            victim.monster.type.friends[aggressor.type.class])
    end
    return true
end


function initialize_states()
    local player

    for player in Players() do
        player._cooldown_timer = 0
        player._autofire = true
        player._track_items = true
        player._track_monsters = true
        player._previous_hotkey = 0
        player._previous_hotkey_tick = 0
    end
end


function valid_target_p(player, target)
    return (
        (is_monster(target) and
            Monsters[target.index] and
            target.visible and
            player.monster ~= target and
            not friendp(player, target) and
            not target.player and
            next_lines[player.polygon.index][target.polygon.index])
        or
        (is_player(target) and
            Players[target.index] and
            player ~= target and
            not target.dead and
            not friendp(player, target) and
            next_lines[player.polygon.index][target.polygon.index])
        or
        (is_item(target) and
            Items[target.index] and
            target.visible and
            player.items[target.type] < target.type.maximum_inventory and
            next_lines[player.polygon.index][target.polygon.index])
    )
end


--[[
Updates a player's target based on idle AI thought.
]]
function refresh_target(player)
    local possible_target

    if player.dead then
        player._target = nil
        return
    end

    -- if our target died, make room for a new one
    if (not valid_target_p(player, player._target)) then
        player._target = nil
    end

    -- if we have a target, no need to update
    if player._target then
        return
    end

    local nearest_distance = 2*1024
    if player._track_monsters then
        -- try monsters
        for possible_target in Monsters() do
            if valid_target_p(player, possible_target) then
                local dx = possible_target.x - player.x
                local dy = possible_target.y - player.y
                if (math.sqrt(dx*dx + dy*dy) < nearest_distance) then
                    nearest_distance = math.sqrt(dx*dx + dy*dy)
                    player._target = possible_target
                end
            end
        end

        -- try players
        for possible_target in Players() do
            if valid_target_p(player, possible_target) then
                local dx = possible_target.x - player.x
                local dy = possible_target.y - player.y
                if (math.sqrt(dx*dx + dy*dy) < nearest_distance) then
                    nearest_distance = math.sqrt(dx*dx + dy*dy)
                    player._target = possible_target
                end
            end
        end
    end

    if player._track_items then
        -- try items
        for possible_target in Items() do
            if valid_target_p(player, possible_target) then
                local dx = possible_target.x - player.x
                local dy = possible_target.y - player.y
                if (math.sqrt(dx*dx + dy*dy) < nearest_distance) then
                    nearest_distance = math.sqrt(dx*dx + dy*dy)
                    player._target = possible_target
                end
            end
        end
    end

    if player._target then
        player._target._is_valid = true
    end
end


function set_overlay_text(player)
    if player.dead then
        player.overlays[0].text = "Dead"
    elseif (player.life < berserk_threshold) then
        player.overlays[0].text = "Berserk"
    elseif (0 < player._cooldown_timer) then
        player.overlays[0].text = "Firing"
    elseif (0 > player._cooldown_timer) then
        player.overlays[0].text = "Cooling"
    else
        if player._target then
            player.overlays[0].text = "Seeking"
        else
            player.overlays[0].text = "Untargeted"
        end
    end

    if is_player(player._target) then
        player.overlays[1].text = player._target.name
    elseif is_monster(player._target) then
        player.overlays[1].text = player._target.type.class.mnemonic
    elseif is_item(player._target) then
        player.overlays[1].text = player._target.type.mnemonic
    else
        player.overlays[1].text = " "
    end

    player.overlays[2].text = "aim: " .. ((player._autofire and "auto") or "user")
    player.overlays[3].text = "monsters: " .. ((player._track_monsters and "on") or "off")
    player.overlays[3].color = ((player._track_monsters and "green") or "red")
    player.overlays[4].text = "items: " .. ((player._track_items and "on") or "off")
    player.overlays[4].color = ((player._track_items and "green") or "red")
end


function update_timers(player)
    if player.dead then
        player._cooldown_timer = 0
    elseif (player.life < berserk_threshold) then
        player._cooldown_timer = Game.ticks + firing_delta
    elseif (0 < player._cooldown_timer) then
        if (Game.ticks > player._cooldown_timer) then
            player._cooldown_timer = -(Game.ticks + cooldown_delta)
        end
    elseif (0 > player._cooldown_timer) then
        if (Game.ticks > -player._cooldown_timer) then
            player._cooldown_timer = 0
        end
    end
end


function update_triggers(player)
    -- player is cooling down, not permitted to fire
    if (player._cooldown_timer < 0) then
        player.action_flags.left_trigger = false
        player.action_flags.right_trigger = false
    end
    -- otherwise, check the trigger state
    if (player.action_flags.left_trigger or player.action_flags.right_trigger) then
        if (player._cooldown_timer == 0) then
            -- player is armed, starting to fire
            player._cooldown_timer = Game.ticks + firing_delta
        end
    end
end


-- TODO: need to rethink this if idle isn't called on laggy ticks
function update_movement(player)
    if player.dead then
        player._old_position = nil
        return
    end

    if nil == player._old_position then
        player._old_position = {
            x = player.x,
            y = player.y,
            z = player.z,
            polygon = player.monster.polygon
        }
    end

    -- always let the engine handle z movement
    player._old_position.z = player.z

    -- if we're untargeted or falling, let player control motion
    if (nil == player._target or
            player.external_velocity.z ~= 0) then
        player._old_position = nil
        return
    end

    --
    -- otherwise, we're going to control player motion.
    --
    
    -- if we're firing and not berserk, stand still
    if (player._cooldown_timer > 0 and player.life > berserk_threshold) then
        player:position(
            player._old_position.x,
            player._old_position.y,
            player._old_position.z,
            player._old_position.polygon.index  -- TODO: is this .index right?
        )
        return
    end

    -- otherwise, track to the target.
    -- calculate the displacement vector...
    local dx, dy = -player.x, -player.y
    local line = next_lines[player.monster.polygon.index][player._target.polygon.index]
    if (line ~= nil and line ~= true) then
        -- otherwise, look up the path and walk to the midpoint
        dx = dx + (line.endpoints[0].x + line.endpoints[1].x) / 2
        dy = dy + (line.endpoints[0].y + line.endpoints[1].y) / 2
    elseif player.monster.polygon == player._target.polygon then
        -- if we're already there, use the targets coordinates directly
        dx, dy = dx + player._target.x, dy + player._target.y
    else
        -- TODO: could do different things here, but we give up
        player._old_position = nil
        player._target = nil
        return
    end

    -- ... normalize it to walking / running speed ...
    local norm = math.sqrt(dx*dx + dy*dy)
    if ((player.monster.polygon == player._target.polygon) and
            (norm < player_thiccness)) then
        -- NOTE: could use polygon:check_collision here to detect object
        --       collisions, but I'm concerned about floor collisions. if i'm on
        --       a cliff and the algorithm wants me to jump off, will that check
        --       collide with the upper cliff floor?
        player._old_position = nil
        return
    end

    -- check that we can step through the line
    if (line and line ~= true and norm < player_thiccness) then
        local other_polygon = (
            (line.cw_polygon == player.polygon
                and line.ccw_polygon)
            or line.cw_polygon
        )
        local my_height = math.max(
            player.monster.z,
            player.monster.polygon.floor.height + player_step
        )
        local target_ceiling = (
            (other_polygon.platform
                and not other_polygon.platform.active
                and other_polygon.platform.ceiling_height)
            or other_polygon.ceiling.height
        )
        local target_floor = (
            (other_polygon.platform
                and not other_polygon.platform.active
                and other_polygon.platform.floor_height)
            or other_polygon.floor.height
        )
        if ((my_height < target_floor)
                or target_ceiling - target_floor < player_height) then
            -- freeze player
            player:position(
                player._old_position.x,
                player._old_position.y,
                player._old_position.z,
                player._old_position.polygon.index
            )

            -- TODO: if the next poly has a player-controllable platform that is
            -- inactive, then activate it.
            if other_polygon.platform and not other_polygon.platform.active then
                player.action_flags.action_trigger = true
            end
            return
        end
    end

    if (player.life < berserk_threshold) then
        norm = norm / berserk_travel_speed
    else
        norm = norm / travel_speed
    end
    dx, dy = dx / norm, dy / norm

    -- ... and calculate the new position vector.
    local new_x, new_y = player._old_position.x + dx, player._old_position.y + dy
    local new_z = player._old_position.z

    -- update which polygon we belong to
    -- NOTE: I tried using polygon:find_polygon here, but it didn't seem to work
    local new_polygon = player._old_position.polygon
    -- if we're able ...
    if nil ~= line and true ~= line then
        -- ... find the "other" polygon
        if player._old_position.polygon == line.cw_polygon then
            new_polygon = line.ccw_polygon
        else
            new_polygon = line.cw_polygon
        end
        -- if we haven't moved to it, then constrain to the old polygon
        if not new_polygon:contains(new_x, new_y) then
            new_polygon = player._old_position.polygon
        end
    end

    -- ... push the player one tick forward
    player:position(new_x, new_y, new_z, new_polygon)
    if (new_polygon ~= player._old_position.polygon  -- remove this if we figure out polygon triggers
            and new_polygon.platform
            and not new_polygon.platform.door
            and new_polygon.platform.player_controllable
            and not new_polygon.platform.active) then
        new_polygon.platform.active = true
    end
    player._old_position = {
        x = new_x,
        y = new_y,
        z = new_z,
        polygon = new_polygon
    }
end


-- NOTE: 0 is east, 90 is south, -90 is north, 180/-180 west
-- NOTE: pitch constrained to [-30, 30], down to up
function update_aiming(player)
    local original_yaw, original_pitch = player.yaw, player.pitch

    -- if we're not hunting, then quit
    if player._target == nil then
        return
    end

    player.action_flags.right_trigger = false
    player.action_flags.left_trigger = false

    -- try aiming at the object and check for visibility
    local dx = player._target.x - player.x
    local dy = player._target.y - player.y
    local dz = player._target.z - player.z
    local desired_yaw = 180 / math.pi * math.atan2(dy, dx)
    local desired_pitch = 180 / math.pi * math.atan2(dz, math.sqrt(dy*dy + dx*dx))

    -- is the target within our FoV?
    player.yaw = desired_yaw
    player.pitch = desired_pitch
    target, tx, ty, tz, tpolygon = player:find_target()
    -- if not...
    if target ~= player._target then
        -- aim for next_lines instead
        local line = next_lines[player.monster.polygon.index][player._target.polygon.index]
        if line ~= nil and line ~= true then
            desired_pitch = 0
            dx = (line.endpoints[0].x + line.endpoints[1].x) / 2 - player.x
            dy = (line.endpoints[0].y + line.endpoints[1].y) / 2 - player.y
            desired_yaw = 180 / math.pi * math.atan2(dy, dx)
        end
    end
    
    -- push the player's pitch and yaw toward the desired direction
    if desired_pitch > original_pitch then
        player.pitch = original_pitch + math.min(desired_pitch - original_pitch, pitch_step)
    elseif desired_pitch < original_pitch then
        player.pitch = original_pitch - math.min(original_pitch - desired_pitch, pitch_step)
    else
        player.pitch = original_pitch
    end

    local centered_yaw_delta = ((desired_yaw - original_yaw + 180) % (2 * 180)) - 180
    if centered_yaw_delta > 0 then
        player.yaw = original_yaw + math.min(yaw_step, centered_yaw_delta)
    elseif centered_yaw_delta < 0 then
        player.yaw = original_yaw - math.min(yaw_step, -centered_yaw_delta)
    else
        player.yaw = desired_yaw
    end

    -- using the actual aim, try to fire
    target, tx, ty, tz, tpolygon = player:find_target()
    if (target == player._target or
            (target.player and target.player == player._target)) then
        player.action_flags.right_trigger = true
        player.action_flags.left_trigger = true
    end
end


function update_hotkeys(player)
    -- too soon to check?
    if player._previous_hotkey_tick + 3 > Game.ticks then
        return
    end
    
    -- debounce
    if player.hotkey == player._previous_hotkey and 0 ~= player._previous_hotkey then
        player._previous_hotkey_tick = Game.ticks
        return
    end

    player._previous_hotkey = player.hotkey

    -- if we saw a hotkey, stash it and wait
    if player.hotkey > 0 then
        player._previous_hotkey_tick = Game.ticks
    end
    
    -- actual hotkey logic
    if player.hotkey == 10 then
        player._autofire = not player._autofire
    elseif player.hotkey == 11 then
        player._track_items = not player._track_items
        player._target = nil
    elseif player.hotkey == 12 then
        player._track_monsters = not player._track_monsters
        player._target = nil
    end
end


-- main body
function Triggers.init(restoring_game)
    precalculate_paths()
    initialize_states()
end

function Triggers.idle()
    for player in Players() do
        update_hotkeys(player)
        update_timers(player)
        refresh_target(player)
        update_movement(player)
        if player._autofire then
            update_aiming(player)
        end
        update_triggers(player)
        set_overlay_text(player)
    end
end

function Triggers.player_damaged(victim, aggressor_player, aggressor_monster,
                                 damage_type, damage_amount, projectile)
    if ((aggressor_player ~= nil) and not friendp(victim, aggressor_player)) then
        victim._target = aggressor_player.monster
    elseif ((aggressor_monster ~= nil) and not friendp(victim, aggressor_monster)) then
        victim._target = aggressor_monster
    end
end
